# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""impact_task tests."""
import unittest

import mock

from clusterfuzz._internal.bot.tasks import impact_task
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.core.bot.tasks.component_revision_patching_test import \
    ComponentRevisionPatchingTest
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


@test_utils.with_cloud_emulators('datastore')
class ExecuteTaskTest(unittest.TestCase):
  """Test execute_task."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.is_chromium',
        'clusterfuzz._internal.bot.tasks.impact_task.get_impacts_from_url',
        'clusterfuzz._internal.bot.tasks.impact_task.get_impacts_on_prod_builds',
        'clusterfuzz._internal.bot.tasks.setup.setup_testcase',
        'clusterfuzz._internal.build_management.build_manager.is_custom_binary',
        'clusterfuzz._internal.build_management.build_manager.has_production_builds',
        'clusterfuzz._internal.bot.testcase_manager.get_command_line_for_application',
        'clusterfuzz._internal.base.tasks.add_task',
    ])
    impacts = impact_task.Impacts(
        stable=impact_task.Impact('stable', False, 'trace-stable'),
        beta=impact_task.Impact('beta', True, 'trace-beta'),
        extended_stable=impact_task.Impact('extended stable', False,
                                           'trace-extended-stable'),
        head=impact_task.Impact('head', False, 'trace-head'))
    self.mock.is_chromium.return_value = True
    self.mock.is_custom_binary.return_value = False
    self.mock.has_production_builds.return_value = True
    self.mock.get_impacts_from_url.return_value = impacts
    self.mock.setup_testcase.return_value = (['a'], None, 'path')
    self.mock.get_impacts_on_prod_builds.return_value = impacts

    self.testcase = data_types.Testcase()
    self.testcase.is_impact_set_flag = False
    self.testcase.status = 'Processed'
    self.testcase.crash_stacktrace = 'trace'
    self.testcase.regression = '123:456'
    self.testcase.job_type = 'job2'
    self.testcase.project_name = 'chromium'
    self.testcase.put()

  def reload(self):
    """Reload testcase."""
    self.testcase = self.testcase.key.get()

  def expect_unchanged(self):
    """Expect testcase's impacts to be unchanged."""
    self.reload()
    self.assertIsNone(self.testcase.impact_stable_version)
    self.assertIsNone(self.testcase.impact_beta_version)
    self.assertIsNone(self.testcase.impact_head_version)

  def expect_changed(self):
    """Expect testcase's impacts to be changed."""
    self.reload()
    self.assertTrue(self.testcase.is_impact_set_flag)
    self.assertEqual('extended stable',
                     self.testcase.impact_extended_stable_version)
    self.assertFalse(self.testcase.impact_stable_version_likely)
    self.assertEqual('stable', self.testcase.impact_stable_version)
    self.assertFalse(self.testcase.impact_stable_version_likely)
    self.assertEqual('beta', self.testcase.impact_beta_version)
    self.assertTrue(self.testcase.impact_beta_version_likely)
    self.assertEqual('head', self.testcase.impact_head_version)
    self.assertFalse(self.testcase.impact_head_version_likely)

  def test_bail_out_non_chromium(self):
    """Test bailing out for non chromium projects."""
    self.mock.is_chromium.return_value = False
    impact_task.execute_task(self.testcase.key.id(), 'job')
    self.expect_unchanged()

  def test_bail_out_fixed(self):
    """Test bailing out when the testcase is fixed."""
    self.testcase.fixed = 'Yes'
    self.testcase.is_impact_set_flag = True
    self.testcase.put()
    impact_task.execute_task(self.testcase.key.id(), 'job')
    self.expect_unchanged()

  def test_bail_out_status_unreproducible(self):
    """Test bailing out when the testcase status is unreproducible (never
    reproduced)."""
    self.testcase.status = 'Unreproducible'
    self.testcase.is_impact_set_flag = True
    self.testcase.put()
    impact_task.execute_task(self.testcase.key.id(), 'job')
    self.expect_unchanged()

  def test_bail_out_custom_binary(self):
    """Test bailing out for custom binary."""
    self.mock.is_custom_binary.return_value = True
    impact_task.execute_task(self.testcase.key.id(), 'job')
    self.expect_unchanged()

  def test_bail_out_unreproducible(self):
    """Test bailing out when the testcase is unreproducible (reproduced once,
    but flaky)."""
    self.testcase.one_time_crasher_flag = True
    self.testcase.put()
    impact_task.execute_task(self.testcase.key.id(), 'job')
    self.expect_unchanged()

  def test_bail_out_non_prod_build_and_no_regression_range(self):
    """Test bailing out when reproducible testcase does not have a regression
    range yet and we dont have production builds to test."""
    self.testcase.one_time_crasher_flag = False
    self.testcase.regression = ''
    self.testcase.put()
    self.mock.has_production_builds.return_value = False
    impact_task.execute_task(self.testcase.key.id(), 'job')
    self.expect_unchanged()

  def test_non_prod_build(self):
    """Test getting impact for non-prod build."""
    self.mock.has_production_builds.return_value = False
    impact_task.execute_task(self.testcase.key.id(), 'job')
    self.expect_changed()
    self.mock.get_impacts_from_url.assert_has_calls(
        [mock.call(self.testcase.regression, self.testcase.job_type)])

  def test_bail_out_setup_testcase(self):
    """Test bailing out when setting up testcase fails."""
    self.mock.has_production_builds.return_value = True
    self.mock.setup_testcase.return_value = ([], None, 'path')
    impact_task.execute_task(self.testcase.key.id(), 'job')
    self.expect_unchanged()

  def test_build_failed_exception(self):
    """Test when BuildFailedException occurs."""
    self.mock.get_impacts_on_prod_builds.side_effect = (
        impact_task.BuildFailedException('error-from-build'))
    impact_task.execute_task(self.testcase.key.id(), 'job')

    self.expect_unchanged()
    self.assertIn('error-from-build', self.testcase.comments)
    self.assertIn(data_types.TaskState.ERROR, self.testcase.comments)
    self.mock.add_task.assert_has_calls(
        [mock.call('impact', self.testcase.key.id(), 'job', wait_time=None)])

  def test_prod_build(self):
    """Test getting impact for prod build."""
    impact_task.execute_task(self.testcase.key.id(), 'job')
    self.expect_changed()
    self.assertIn(data_types.TaskState.FINISHED, self.testcase.comments)
    self.assertNotIn('trace-stable', self.testcase.crash_stacktrace)
    self.assertNotIn('trace-beta', self.testcase.crash_stacktrace)
    self.mock.get_impacts_on_prod_builds.assert_has_calls(
        [mock.call(mock.ANY, 'path')])

  def test_prod_build_unreproducible(self):
    """Test getting impact for prod build (unreproducible)."""
    self.testcase.status = 'Unreproducible'
    self.testcase.put()
    impact_task.execute_task(self.testcase.key.id(), 'job')

    self.expect_changed()
    self.assertIn(data_types.TaskState.FINISHED, self.testcase.comments)
    self.assertIn('trace-stable', self.testcase.crash_stacktrace)
    self.assertIn('trace-beta', self.testcase.crash_stacktrace)
    self.mock.get_impacts_on_prod_builds.assert_has_calls(
        [mock.call(mock.ANY, 'path')])


class GetImpactsFromUrlTest(ComponentRevisionPatchingTest):
  """Test get_impacts_from_url."""

  def setUp(self):
    """Setup for get impacts from url test."""
    super().setUp()
    helpers.patch(self, [
        'clusterfuzz._internal.bot.tasks.impact_task.get_start_and_end_revision',
        'clusterfuzz._internal.bot.tasks.impact_task.get_impact',
        'clusterfuzz._internal.chrome.build_info.get_build_to_revision_mappings',
        'clusterfuzz._internal.build_management.revisions.revision_to_branched_from',
        'clusterfuzz._internal.datastore.data_handler.get_component_name',
    ])
    self.mock.get_component_name.return_value = None
    self.mock.revision_to_branched_from.side_effect = (
        self.mock_revision_to_branched_from)
    self.mock.get_start_and_end_revision.return_value = (1, 100)
    self.mock.get_build_to_revision_mappings.return_value = {
        'stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'beta': {
            'revision': '399171',
            'version': '75.0.1353.43'
        },
        'canary': {
            'revision': '400000',
            'version': '76.0.1234.43'
        }
    }
    self.mock.get_impact.side_effect = [
        impact_task.Impact(),
        impact_task.Impact('s', False),
        impact_task.Impact('b', True),
        impact_task.Impact('c', False)
    ]

  @staticmethod
  def mock_revision_to_branched_from(uri, rev):
    if uri == 'fish':
      return None
    if rev == 'cba1fdd4d72e7c5b874f9eeb07901792f26c871a':
      return '666666'
    if rev == '3a590058de9b3640f73741b1e95f815f5c089988':
      return '888888'
    return '777777'

  def test_bail_out_unknown_component(self):
    """Test bailing out when having an unknown component."""
    self.mock.get_component_name.return_value = 'com'
    self.assertTrue(
        impact_task.get_impacts_from_url('123:456', 'job').is_empty())
    self.mock.get_start_and_end_revision.assert_has_calls([])
    self.mock.get_build_to_revision_mappings.assert_has_calls([])
    self.mock.get_impact.assert_has_calls([])

  def test_bail_out_end_revision(self):
    """Test bailing out when there's no end_revision."""
    self.mock.get_start_and_end_revision.return_value = (1, None)
    self.assertTrue(
        impact_task.get_impacts_from_url('123:456', 'job').is_empty())
    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls([])
    self.mock.get_impact.assert_has_calls([])

  def test_bail_out_build_revision_mappings(self):
    """Test bailing out when there's no mapping."""
    self.mock.get_build_to_revision_mappings.return_value = {}
    self.assertTrue(
        impact_task.get_impacts_from_url('123:456', 'job',
                                         'windows').is_empty())
    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])
    self.mock.get_impact.assert_has_calls([])

  def test_bail_out_no_build_to_revision_mapping(self):
    """Test bailing out when get_build_to_revision_mapping is empty"""
    self.mock.get_build_to_revision_mappings.return_value = None
    self.assertTrue(
        impact_task.get_impacts_from_url('123:456', 'job',
                                         'windows').is_empty())
    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])
    self.mock.get_impact.assert_has_calls([])

  def test_bail_out_no_component_branched_from(self):
    """Test bailing out when there's no Cr-Branched-From."""
    self.mock.get_component_name.return_value = 'fish'
    self.assertTrue(
        impact_task.get_impacts_from_url('123:456', 'job',
                                         'windows').is_empty())
    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])

  def test_bail_if_two_identically_named_components(self):
    """Tests we bail if a comp is given twice in the component deps."""
    self.mock.get_build_to_revision_mappings.return_value = {
        'stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'beta': {
            'revision': '400000',
            'version': '76.0.1353.43'
        }
    }
    self.mock.get_component_name.return_value = 'skia'
    self.assertTrue(
        impact_task.get_impacts_from_url('123:456', 'job',
                                         'windows').is_empty())
    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])

  def test_get_impacts_es_not_exists(self):
    """Test getting impacts when extended stable doesn't exist."""
    impacts = impact_task.get_impacts_from_url('123:456', 'job', 'windows')

    self.assertEqual('', impacts.extended_stable.version)
    self.assertFalse(impacts.extended_stable.likely)
    self.assertEqual('s', impacts.stable.version)
    self.assertFalse(impacts.stable.likely)
    self.assertEqual('b', impacts.beta.version)
    self.assertTrue(impacts.beta.likely)
    self.assertEqual('c', impacts.head.version)
    self.assertFalse(impacts.head.likely)

    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])
    self.mock.get_impact.assert_has_calls([
        mock.call(None, 1, 100),
        mock.call({
            'version': '74.0.1345.34',
            'revision': '398287'
        }, 1, 100),
        mock.call({
            'version': '75.0.1353.43',
            'revision': '399171'
        }, 1, 100)
    ])

  def test_get_impacts_es_exists(self):
    """Test getting impacts when extended stable exists."""
    self.mock.get_build_to_revision_mappings.return_value = {
        'extended_stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'beta': {
            'revision': '399171',
            'version': '75.0.1353.43'
        },
        'canary': {
            'revision': '400000',
            'version': '76.0.1234.43'
        }
    }
    self.mock.get_impact.side_effect = [
        impact_task.Impact('es', False),
        impact_task.Impact('s', False),
        impact_task.Impact('b', True),
        impact_task.Impact('c', False)
    ]

    impacts = impact_task.get_impacts_from_url('123:456', 'job', 'windows')

    self.assertEqual('es', impacts.extended_stable.version)
    self.assertFalse(impacts.extended_stable.likely)
    self.assertEqual('s', impacts.stable.version)
    self.assertFalse(impacts.stable.likely)
    self.assertEqual('b', impacts.beta.version)
    self.assertTrue(impacts.beta.likely)
    self.assertEqual('c', impacts.head.version)
    self.assertFalse(impacts.head.likely)

    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])
    self.mock.get_impact.assert_has_calls([
        mock.call({
            'version': '74.0.1345.34',
            'revision': '398287'
        }, 1, 100),
        mock.call({
            'version': '74.0.1345.34',
            'revision': '398287'
        }, 1, 100),
        mock.call({
            'version': '75.0.1353.43',
            'revision': '399171'
        }, 1, 100),
        mock.call({
            'version': '76.0.1234.43',
            'revision': '400000'
        }, 1, 100, True)
    ])

  def test_get_impacts_canary_not_exists(self):
    """Test getting impacts when extended stable exists."""
    self.mock.get_build_to_revision_mappings.return_value = {
        'extended_stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'beta': {
            'revision': '399171',
            'version': '75.0.1353.43'
        },
        'dev': {
            'revision': '400000',
            'version': '76.0.1234.43'
        }
    }
    self.mock.get_impact.side_effect = [
        impact_task.Impact('es', False),
        impact_task.Impact('s', False),
        impact_task.Impact('b', True),
        impact_task.Impact('d', False)
    ]

    impacts = impact_task.get_impacts_from_url('123:456', 'job', 'windows')

    self.assertEqual('es', impacts.extended_stable.version)
    self.assertFalse(impacts.extended_stable.likely)
    self.assertEqual('s', impacts.stable.version)
    self.assertFalse(impacts.stable.likely)
    self.assertEqual('b', impacts.beta.version)
    self.assertTrue(impacts.beta.likely)
    self.assertEqual('d', impacts.head.version)
    self.assertFalse(impacts.head.likely)

    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])
    self.mock.get_impact.assert_has_calls([
        mock.call({
            'version': '74.0.1345.34',
            'revision': '398287'
        }, 1, 100),
        mock.call({
            'version': '74.0.1345.34',
            'revision': '398287'
        }, 1, 100),
        mock.call({
            'version': '75.0.1353.43',
            'revision': '399171'
        }, 1, 100),
        mock.call({
            'version': '76.0.1234.43',
            'revision': '400000'
        }, 1, 100, True)
    ])

  def test_get_impacts_known_component_es_not_exists(self):
    """Test getting impacts for a known component
    when extended stable doesn't exists."""
    self.mock.get_component_name.return_value = 'v8'
    self.mock.get_impact.side_effect = [
        impact_task.Impact('s', False),
        impact_task.Impact('b', True),
        impact_task.Impact('c', False)
    ]
    impacts = impact_task.get_impacts_from_url('123:456', 'job', 'windows')

    self.assertEqual('', impacts.extended_stable.version)
    self.assertFalse(impacts.extended_stable.likely)
    self.assertEqual('s', impacts.stable.version)
    self.assertFalse(impacts.stable.likely)
    self.assertEqual('b', impacts.beta.version)
    self.assertTrue(impacts.beta.likely)

    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])
    self.mock.get_impact.assert_has_calls([
        mock.call({
            'version': '74.0.1345.34',
            'revision': '666666'
        }, 1, 100, False),
        mock.call({
            'version': '75.0.1353.43',
            'revision': '888888'
        }, 1, 100, False),
        mock.call({
            'version': '76.0.1234.43',
            'revision': '888888'
        }, 1, 100, True)
    ])

  def test_get_impacts_known_component_es_exists(self):
    """Test getting impacts for a known component
    when extended stable exists."""
    self.mock.get_component_name.return_value = 'v8'
    self.mock.get_build_to_revision_mappings.return_value = {
        'extended_stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'beta': {
            'revision': '399171',
            'version': '75.0.1353.43'
        },
        'canary': {
            'revision': '400000',
            'version': '76.0.1234.43'
        }
    }
    self.mock.get_impact.side_effect = [
        impact_task.Impact('es', False),
        impact_task.Impact('s', False),
        impact_task.Impact('b', True),
        impact_task.Impact('c', True)
    ]
    impacts = impact_task.get_impacts_from_url('123:456', 'job', 'windows')

    self.assertEqual('es', impacts.extended_stable.version)
    self.assertFalse(impacts.extended_stable.likely)
    self.assertEqual('s', impacts.stable.version)
    self.assertFalse(impacts.stable.likely)
    self.assertEqual('b', impacts.beta.version)
    self.assertTrue(impacts.beta.likely)
    self.assertEqual('c', impacts.head.version)
    self.assertTrue(impacts.head.likely)

    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])
    self.mock.get_impact.assert_has_calls([
        mock.call({
            'version': '74.0.1345.34',
            'revision': '666666'
        }, 1, 100, False),
        mock.call({
            'version': '75.0.1353.43',
            'revision': '888888'
        }, 1, 100, False),
        mock.call({
            'version': '76.0.1234.43',
            'revision': '888888'
        }, 1, 100, True)
    ])

  def test_get_impacts_known_component_es_exists_canary_not_exists(self):
    """Test getting impacts for a known component
    when extended stable exists and canary doesn't."""
    self.mock.get_component_name.return_value = 'v8'
    self.mock.get_build_to_revision_mappings.return_value = {
        'extended_stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'beta': {
            'revision': '399171',
            'version': '75.0.1353.43'
        },
        'dev': {
            'revision': '400000',
            'version': '76.0.1234.43'
        }
    }
    self.mock.get_impact.side_effect = [
        impact_task.Impact('es', False),
        impact_task.Impact('s', False),
        impact_task.Impact('b', True),
        impact_task.Impact('c', True)
    ]
    impacts = impact_task.get_impacts_from_url('123:456', 'job', 'windows')

    self.assertEqual('es', impacts.extended_stable.version)
    self.assertFalse(impacts.extended_stable.likely)
    self.assertEqual('s', impacts.stable.version)
    self.assertFalse(impacts.stable.likely)
    self.assertEqual('b', impacts.beta.version)
    self.assertTrue(impacts.beta.likely)
    self.assertEqual('c', impacts.head.version)
    self.assertTrue(impacts.head.likely)

    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456', 'job')])
    self.mock.get_build_to_revision_mappings.assert_has_calls(
        [mock.call('windows')])
    self.mock.get_impact.assert_has_calls([
        mock.call({
            'version': '74.0.1345.34',
            'revision': '666666'
        }, 1, 100, False),
        mock.call({
            'version': '75.0.1353.43',
            'revision': '888888'
        }, 1, 100, False),
        mock.call({
            'version': '76.0.1234.43',
            'revision': '888888'
        }, 1, 100, True)
    ])


class GetImpactTest(unittest.TestCase):
  """Test get_impact."""

  def test_bail_out_build_revision(self):
    """Test bailing out when there's no build_revision."""
    self.assertTrue(impact_task.get_impact({}, 1, 100).is_empty())

  def test_bail_out_non_digit(self):
    """Test bailing out when revision is not a number."""
    self.assertTrue(
        impact_task.get_impact({
            'revision': 'aa'
        }, 1, 100).is_empty())

  def test_empty_impact(self):
    """Test returning empty impact when start_revision is greater than the build
      revision."""
    self.assertTrue(
        impact_task.get_impact({
            'revision': '10',
            'version': '50'
        }, 20, 100).is_empty())

  def test_get_impact(self):
    """Test getting version."""
    impact = impact_task.get_impact({'revision': '30', 'version': '50'}, 20, 25)
    self.assertEqual('50', impact.version)
    self.assertFalse(impact.likely)
    self.assertEqual('', impact.extra_trace)

  def test_get_likely_impact(self):
    """Test getting likely version."""
    impact = impact_task.get_impact({'revision': '30', 'version': '50'}, 20, 31)
    self.assertEqual('50', impact.version)
    self.assertTrue(impact.likely)
    self.assertEqual('', impact.extra_trace)

  def test_get_beyond_build(self):
    """Test if the regression range is beyond the version."""
    impact = impact_task.get_impact({'revision': '30', 'version': '50'}, 31, 32)
    self.assertTrue(impact.is_empty())

  def test_get_beyond_build_if_final(self):
    """Test getting likely version."""
    impact = impact_task.get_impact({
        'revision': '30',
        'version': '50.1.2.3'
    }, 31, 32, True)
    self.assertEqual('50', impact.version)
    self.assertTrue(impact.likely)
    self.assertEqual('', impact.extra_trace)


class GetImpactsOnProdBuilds(unittest.TestCase):
  """Test get_impacts_on_prod_builds."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.bot.tasks.impact_task.get_impact_on_build',
        'clusterfuzz._internal.bot.tasks.impact_task.get_impacts_from_url',
        'clusterfuzz._internal.bot.testcase_manager.get_command_line_for_application',
        'clusterfuzz._internal.chrome.build_info.get_build_to_revision_mappings',
    ])
    self.mock.get_build_to_revision_mappings.return_value = {
        'stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'beta': {
            'revision': '399171',
            'version': '75.0.1353.43'
        },
        'canary': {
            'revision': '400000',
            'version': '76.0.1234.43'
        }
    }
    self.impacts = impact_task.Impacts(
        stable=impact_task.Impact('s', False),
        beta=impact_task.Impact('b', True),
        extended_stable=impact_task.Impact('es', False),
        head=impact_task.Impact('76.0.1234.43', False))

    self.testcase = data_types.Testcase()
    self.testcase.job_type = 'job'
    self.testcase.impact_extended_stable_version = 'es-ver'
    self.testcase.impact_stable_version = 's-ver'
    self.testcase.impact_beta_version = 'b-ver'
    self.testcase.regression = '123:456'

  def test_app_failed_on_stable(self):
    """Test raising AppFailedException when getting stable impact."""
    self.mock.get_impact_on_build.side_effect = impact_task.AppFailedException()
    self.mock.get_impacts_from_url.return_value = self.impacts

    self.assertEqual(
        self.impacts,
        impact_task.get_impacts_on_prod_builds(self.testcase, 'path'))
    self.mock.get_impact_on_build.assert_has_calls([
        mock.call('stable', self.testcase.impact_stable_version, self.testcase,
                  'path')
    ])
    self.mock.get_impacts_from_url.assert_has_calls(
        [mock.call(self.testcase.regression, self.testcase.job_type)])

  def test_app_failed_on_beta(self):
    """Test app fail on beta."""
    self.mock.get_impact_on_build.side_effect = [
        self.impacts.stable,
        impact_task.AppFailedException(),
        self.impacts.extended_stable,
    ]

    self.assertEqual(
        impact_task.Impacts(
            self.impacts.stable,
            impact_task.Impact(),
            self.impacts.extended_stable,
            head=self.impacts.head),
        impact_task.get_impacts_on_prod_builds(self.testcase, 'path'))
    self.mock.get_impact_on_build.assert_has_calls([
        mock.call('stable', self.testcase.impact_stable_version, self.testcase,
                  'path'),
        mock.call('beta', self.testcase.impact_beta_version, self.testcase,
                  'path'),
        mock.call('extended_stable',
                  self.testcase.impact_extended_stable_version, self.testcase,
                  'path'),
    ])
    self.mock.get_impacts_from_url.assert_has_calls([])

  def test_app_failed_on_extended_stable(self):
    """Test app fail on extended stable."""
    self.mock.get_impact_on_build.side_effect = [
        self.impacts.stable,
        self.impacts.beta,
        impact_task.AppFailedException(),
    ]
    self.mock.get_impacts_from_url.return_value = self.impacts

    self.assertEqual(
        self.impacts,
        impact_task.get_impacts_on_prod_builds(self.testcase, 'path'))
    self.mock.get_impact_on_build.assert_has_calls([
        mock.call('stable', self.testcase.impact_stable_version, self.testcase,
                  'path'),
        mock.call('beta', self.testcase.impact_beta_version, self.testcase,
                  'path'),
        mock.call('extended_stable',
                  self.testcase.impact_extended_stable_version, self.testcase,
                  'path'),
    ])
    self.mock.get_impacts_from_url.assert_has_calls(
        [mock.call(self.testcase.regression, self.testcase.job_type)])

  def test_get_impacts(self):
    """Test getting impacts."""
    self.mock.get_impact_on_build.side_effect = [
        self.impacts.stable, self.impacts.beta, self.impacts.extended_stable
    ]

    self.assertEqual(
        self.impacts,
        impact_task.get_impacts_on_prod_builds(self.testcase, 'path'))
    self.mock.get_impact_on_build.assert_has_calls([
        mock.call('stable', self.testcase.impact_stable_version, self.testcase,
                  'path'),
        mock.call('beta', self.testcase.impact_beta_version, self.testcase,
                  'path'),
        mock.call('extended_stable',
                  self.testcase.impact_extended_stable_version, self.testcase,
                  'path'),
    ])
    self.mock.get_impacts_from_url.assert_has_calls([])

  def test_get_impacts_canary_not_exists(self):
    """Test getting impacts if there's no Canary build, only dev."""
    self.mock.get_impact_on_build.side_effect = [
        self.impacts.stable, self.impacts.beta, self.impacts.extended_stable
    ]
    self.mock.get_build_to_revision_mappings.return_value = {
        'stable': {
            'revision': '398287',
            'version': '74.0.1345.34'
        },
        'beta': {
            'revision': '399171',
            'version': '75.0.1353.43'
        },
        'dev': {
            'revision': '400000',
            'version': '76.0.1234.43'
        }
    }

    self.assertEqual(
        self.impacts,
        impact_task.get_impacts_on_prod_builds(self.testcase, 'path'))
    self.mock.get_impact_on_build.assert_has_calls([
        mock.call('stable', self.testcase.impact_stable_version, self.testcase,
                  'path'),
        mock.call('beta', self.testcase.impact_beta_version, self.testcase,
                  'path'),
        mock.call('extended_stable',
                  self.testcase.impact_extended_stable_version, self.testcase,
                  'path'),
    ])
    self.mock.get_impacts_from_url.assert_has_calls([])


class GetImpactOnBuild(unittest.TestCase):
  """Test get_impact_on_build."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.build_management.build_manager.setup_production_build',
        'clusterfuzz._internal.system.environment.get_value',
        'clusterfuzz._internal.bot.testcase_manager.get_command_line_for_application',
        'clusterfuzz._internal.bot.testcase_manager.test_for_crash_with_retries',
    ])
    self.env = {
        'APP_PATH': 'app',
        'TEST_TIMEOUT': '9',
        'JOB_NAME': 'linux_asan_chrome',
        'TOOL_NAME': 'ASAN'
    }
    self.result = mock.Mock()
    self.result.is_crash = mock.Mock()
    self.result.get_stacktrace = mock.Mock()

    self.mock.setup_production_build.return_value = (
        build_manager.ProductionBuild('/base', '52', None, 'stable'))
    self.mock.get_value.side_effect = self.env.get
    self.mock.test_for_crash_with_retries.return_value = self.result
    self.result.is_crash.return_value = True
    self.result.get_stacktrace.return_value = 'crashed-trace'

    self.testcase = data_types.Testcase()

  def test_build_failed(self):
    """Test raising BuildFailedException."""
    self.mock.setup_production_build.return_value = None
    with self.assertRaises(impact_task.BuildFailedException) as cm:
      impact_task.get_impact_on_build('stable', '52', self.testcase, 'path')

    self.assertEqual('Build setup failed for Stable', str(cm.exception))

  def test_app_failed(self):
    """Test raising AppFailedException."""
    self.env['APP_PATH'] = ''
    self.env['APP_NAME'] = 'app_name'
    with self.assertRaises(impact_task.AppFailedException):
      impact_task.get_impact_on_build('stable', '52', self.testcase, 'path')

  def test_same_version(self):
    """Test same version."""
    impact = impact_task.get_impact_on_build('stable', '52', self.testcase,
                                             'path')
    self.assertEqual('52', impact.version)
    self.assertFalse(impact.likely)
    self.assertEqual('', impact.extra_trace)

  def test_crash(self):
    """Test crashing."""
    impact = impact_task.get_impact_on_build('stable', '53', self.testcase,
                                             'path')
    self.assertEqual('52', impact.version)
    self.assertFalse(impact.likely)
    self.assertIn('crashed-trace', impact.extra_trace)

  def test_not_crash(self):
    """Test not crashing and returning empty impact."""
    self.result.is_crash.return_value = False
    self.assertEqual(
        impact_task.Impact(),
        impact_task.get_impact_on_build('stable', '53', self.testcase, 'path'))


class GetStartAndEndRevisionTest(unittest.TestCase):
  """Test get_start_and_end_revision."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.build_management.revisions.get_start_and_end_revision',
        'clusterfuzz._internal.build_management.revisions.get_component_range_list',
        'clusterfuzz._internal.bot.testcase_manager.get_command_line_for_application',
        'clusterfuzz._internal.system.environment.is_android',
    ])

  def test_normal(self):
    """Test when there's no end revision."""
    self.mock.is_android.return_value = False
    self.mock.get_start_and_end_revision.return_value = (1, 100)
    start, end = impact_task.get_start_and_end_revision('123:456', 'job')

    self.assertEqual(1, start)
    self.assertEqual(100, end)
    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456')])
    self.mock.get_component_range_list.assert_has_calls([])

  def test_android(self):
    """Test android."""
    self.mock.is_android.return_value = True
    self.mock.get_start_and_end_revision.side_effect = [(1, 100), (9, 90)]
    self.mock.get_component_range_list.return_value = [{
        'component': 'test'
    }, {
        'component': 'Chromium',
        'link_text': 'somelink'
    }, {
        'component': 'test'
    }]
    start, end = impact_task.get_start_and_end_revision('123:456',
                                                        'android_job')

    self.assertEqual(9, start)
    self.assertEqual(90, end)
    self.mock.get_start_and_end_revision.assert_has_calls(
        [mock.call('123:456'), mock.call('somelink')])
    self.mock.get_component_range_list.assert_has_calls(
        [mock.call(1, 100, 'android_job')])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""deploy.py handles the deploy command"""

from collections import namedtuple
import contextlib
import datetime
import json
import os
import re
import sys
import tempfile
import time

from local.butler import appengine
from local.butler import common
from local.butler import constants
from local.butler import package
from src.clusterfuzz._internal.config import local_config
from src.clusterfuzz._internal.system import environment

EXPECTED_BOT_COUNT_PERCENT = 0.8

APPENGINE_FILESIZE_LIMIT = 30 * 1000 * 1000  # ~32 MB
DEPLOY_RETRIES = 3
MATCH_ALL = '*'
RETRY_WAIT_SECONDS = 10

# Give 12 hours for cron jobs to complete before deleting a version.
VERSION_DELETE_WINDOW_MINUTES = 12 * 60

INDEX_YAML_PATH = os.path.join(appengine.SRC_DIR_PY, 'index.yaml')
SERVICE_REGEX = re.compile(r'service\s*:\s*(.*)')

Version = namedtuple('Version', ['id', 'deploy_time', 'traffic_split'])


def now():
  """Used for mocks."""
  return datetime.datetime.now()


def _get_services(paths):
  """Get list of services from deployment yamls."""
  services = []
  for path in paths:
    for line in open(path):
      match = SERVICE_REGEX.search(line)
      if match:
        matched_service = match.group(1)
        if matched_service not in services:
          services.append(matched_service)
        break

  return services


def _get_redis_ip(project):
  """Get the redis IP address."""
  region = appengine.region(project)
  return_code, ip = common.execute(
      'gcloud redis instances describe redis-instance '
      '--project={project} --region={region} '
      '--format="value(host)"'.format(project=project, region=region))

  if return_code:
    raise RuntimeError('Failed to get redis IP.')

  return ip.decode('utf-8').strip()


def _additional_app_env_vars(project):
  """Additional environment variables to include for App Engine."""
  return {
      'REDIS_HOST': _get_redis_ip(project),
  }


def _deploy_app_prod(project,
                     deployment_bucket,
                     yaml_paths,
                     package_zip_paths,
                     deploy_appengine=True):
  """Deploy app in production."""
  if deploy_appengine:
    services = _get_services(yaml_paths)
    rebased_yaml_paths = appengine.copy_yamls_and_preprocess(
        yaml_paths, _additional_app_env_vars(project))

    _deploy_appengine(
        project, [INDEX_YAML_PATH] + rebased_yaml_paths,
        stop_previous_version=False)
    for path in rebased_yaml_paths:
      os.remove(path)

    for service in services:
      _delete_old_versions(project, service, VERSION_DELETE_WINDOW_MINUTES)

  if package_zip_paths:
    for package_zip_path in package_zip_paths:
      _deploy_zip(deployment_bucket, package_zip_path)

    _deploy_manifest(deployment_bucket, constants.PACKAGE_TARGET_MANIFEST_PATH)


def _deploy_app_staging(project, yaml_paths):
  """Deploy app in staging."""
  services = _get_services(yaml_paths)

  rebased_yaml_paths = appengine.copy_yamls_and_preprocess(
      yaml_paths, _additional_app_env_vars(project))
  _deploy_appengine(project, rebased_yaml_paths, stop_previous_version=True)
  for path in rebased_yaml_paths:
    os.remove(path)

  for service in services:
    _delete_old_versions(project, service, 0)


def _versions_to_delete(versions, window):
  """Return the versions that should be deleted."""
  # gcloud app versions list returns local time.
  cutoff = now() - datetime.timedelta(minutes=window)

  # Don't delete any versions that stopped serving within
  # |window| minutes before now (or the latest one, since that's definitely
  # still serving).
  # This is so that cron jobs have a chance to finish.

  # Find the first version for which the deploy time of the next version is
  # after the cutoff. This is the first version that we do not delete, because
  # it was still serving after the cutoff.
  delete_end = 0
  while (delete_end < len(versions) - 1 and
         versions[delete_end + 1].deploy_time <= cutoff):
    delete_end += 1

  return versions[:delete_end]


def _delete_old_versions(project, service, delete_window):
  """Delete old versions."""

  def _to_datetime(entry):
    """Parse datetime entry."""
    return datetime.datetime(entry['year'], entry['month'], entry['day'],
                             entry['hour'], entry['minute'], entry['second'])

  _, versions = common.execute('gcloud app versions list --format=json '
                               '--project=%s --service=%s' % (project, service))
  versions = [
      Version(version['id'], _to_datetime(version['last_deployed_time']),
              version['traffic_split']) for version in json.loads(versions)
  ]

  versions.sort(key=lambda v: v.deploy_time)
  assert versions[-1].traffic_split == 1.0

  to_delete = _versions_to_delete(versions, delete_window)
  if not to_delete:
    return

  versions = ' '.join(version.id for version in to_delete)
  common.execute('gcloud app versions delete --quiet '
                 '--project=%s --service=%s %s' % (project, service, versions))


def _deploy_appengine(project, yamls, stop_previous_version, version=None):
  """Deploy to appengine using `yamls`."""
  stop_previous_version_arg = ('--stop-previous-version'
                               if stop_previous_version else
                               '--no-stop-previous-version')

  version_arg = '--version=' + version if version else ''

  for retry_num in range(DEPLOY_RETRIES + 1):
    return_code, _ = common.execute(
        'gcloud app deploy %s --quiet '
        '--project=%s %s %s' % (stop_previous_version_arg, project, version_arg,
                                ' '.join(yamls)),
        exit_on_error=False)

    if return_code == 0:
      break

    if retry_num == DEPLOY_RETRIES:
      print('Failed to deploy after %d retries.' % DEPLOY_RETRIES)
      sys.exit(return_code)

    print('gcloud deployment failed, retrying...')
    time.sleep(RETRY_WAIT_SECONDS)


def find_file_exceeding_limit(path, limit):
  """Find one individual file that exceeds limit within path (recursively)."""
  for root, _, filenames in os.walk(path):
    for filename in filenames:
      full_path = os.path.join(root, filename)
      if os.path.getsize(full_path) >= limit:
        return full_path
  return None


def _deploy_zip(bucket_name, zip_path):
  """Deploy zip to GCS."""
  common.execute('gsutil cp %s gs://%s/%s' % (zip_path, bucket_name,
                                              os.path.basename(zip_path)))


def _deploy_manifest(bucket_name, manifest_path):
  """Deploy source manifest to GCS."""
  if sys.version_info.major == 3:
    manifest_suffix = '.3'
  else:
    manifest_suffix = ''

  common.execute('gsutil cp -a public-read %s '
                 'gs://%s/clusterfuzz-source.manifest%s' %
                 (manifest_path, bucket_name, manifest_suffix))


def _update_deployment_manager(project, name, config_path):
  """Update deployment manager settings."""
  if not os.path.exists(config_path):
    return

  gcloud = common.Gcloud(project)
  operation = 'update'
  try:
    gcloud.run('deployment-manager', 'deployments', 'describe', name)
  except common.GcloudError:
    # Does not exist.
    operation = 'create'

  for _ in range(DEPLOY_RETRIES + 1):
    try:
      gcloud.run('deployment-manager', 'deployments', operation, name,
                 '--config=' + config_path)
      break
    except common.GcloudError:
      time.sleep(RETRY_WAIT_SECONDS)


def _update_pubsub_queues(project):
  """Update pubsub queues."""
  _update_deployment_manager(
      project, 'pubsub',
      os.path.join(environment.get_config_directory(), 'pubsub', 'queues.yaml'))


def _get_region_counts():
  """Get region instance counts."""
  counts = {}
  regions = local_config.MonitoringRegionsConfig()
  clusters = local_config.Config(local_config.GCE_CLUSTERS_PATH).get()

  def get_region(name):
    """Get the region."""
    for pattern in regions.get('patterns'):
      if re.match(pattern['pattern'], name + '-0000'):
        return pattern['name']

    return None

  # Compute expected bot counts per region.
  for config in clusters.values():
    for name, cluster in config['clusters'].items():
      region = get_region(name)
      if not region:
        continue

      counts.setdefault(region, 0)
      counts[region] += cluster['instance_count']

  return counts


@contextlib.contextmanager
def _preprocess_alerts(alerts_path):
  """Preprocess alerts."""
  with open(alerts_path) as f:
    alerts_data = f.read()

  counts = _get_region_counts()
  for region, count in counts.items():
    alerts_data = re.sub('BOT_COUNT:' + region + r'(?=\s|$)',
                         str(int(count * EXPECTED_BOT_COUNT_PERCENT)),
                         alerts_data)

  with tempfile.NamedTemporaryFile(mode='w') as f:
    f.write(alerts_data)
    f.flush()
    yield f.name


def _update_alerts(project):
  """Update pubsub topics."""
  if not local_config.ProjectConfig().get('monitoring.enabled'):
    return

  alerts_path = os.path.join(environment.get_config_directory(), 'monitoring',
                             'alerts.yaml')
  with _preprocess_alerts(alerts_path) as processed_alerts_path:
    _update_deployment_manager(project, 'alerts', processed_alerts_path)


def _update_bigquery(project):
  """Update bigquery datasets and tables."""
  _update_deployment_manager(
      project, 'bigquery',
      os.path.join(environment.get_config_directory(), 'bigquery',
                   'datasets.yaml'))


def _update_redis(project):
  """Update redis instance."""
  _update_deployment_manager(
      project, 'redis',
      os.path.join(environment.get_config_directory(), 'redis',
                   'instance.yaml'))

  region = appengine.region(project)
  return_code, _ = common.execute(
      'gcloud compute networks vpc-access connectors describe '
      'connector --region={region} '
      '--project={project}'.format(project=project, region=region),
      exit_on_error=False)

  if return_code:
    # Does not exist.
    common.execute('gcloud compute networks vpc-access connectors create '
                   'connector --network=default --region={region} '
                   '--range=10.8.0.0/28 '
                   '--project={project}'.format(project=project, region=region))


def get_remote_sha():
  """Get remote sha of origin/master."""
  _, remote_sha_line = common.execute('git ls-remote origin refs/heads/master')

  return re.split(br'\s+', remote_sha_line)[0]


def is_diff_origin_master():
  """Check if the current state is different from origin/master."""
  common.execute('git fetch')
  remote_sha = get_remote_sha()
  _, local_sha = common.execute('git rev-parse HEAD')
  _, diff_output = common.execute('git diff origin/master --stat')

  return diff_output.strip() or remote_sha.strip() != local_sha.strip()


def _staging_deployment_helper(python3=True):
  """Helper for staging deployment."""
  config = local_config.Config(local_config.GAE_CONFIG_PATH)
  project = config.get('application_id')

  print('Deploying %s to staging.' % project)
  deployment_config = config.sub_config('deployment')

  if python3:
    path = 'staging3'
  else:
    path = 'staging'

  yaml_paths = deployment_config.get_absolute_path(path)

  _deploy_app_staging(project, yaml_paths)
  print('Staging deployment finished.')


def _prod_deployment_helper(config_dir,
                            package_zip_paths,
                            deploy_appengine=True,
                            python3=True):
  """Helper for production deployment."""
  config = local_config.Config()
  deployment_bucket = config.get('project.deployment.bucket')

  gae_config = config.sub_config(local_config.GAE_CONFIG_PATH)
  gae_deployment = gae_config.sub_config('deployment')
  project = gae_config.get('application_id')

  print('Deploying %s to prod.' % project)
  if python3:
    path = 'prod3'
  else:
    path = 'prod'

  yaml_paths = gae_deployment.get_absolute_path(path, default=[])
  if not yaml_paths:
    deploy_appengine = False

  if deploy_appengine:
    _update_pubsub_queues(project)
    _update_alerts(project)
    _update_bigquery(project)
    _update_redis(project)

  _deploy_app_prod(
      project,
      deployment_bucket,
      yaml_paths,
      package_zip_paths,
      deploy_appengine=deploy_appengine)

  if deploy_appengine:
    common.execute('python butler.py run setup --config-dir {config_dir} '
                   '--non-dry-run'.format(config_dir=config_dir))
  print('Production deployment finished.')


def execute(args):
  """Deploy Clusterfuzz to Appengine."""
  if sys.version_info.major != 3 or sys.version_info.minor != 7:
    print('You can only deploy from Python 3.7. Install Python 3.7 and '
          'run: `PYTHON=python3.7 local/install_deps.bash`')
    sys.exit(1)

  os.environ['ROOT_DIR'] = '.'

  if not os.path.exists(args.config_dir):
    print('Please provide a valid configuration directory.')
    sys.exit(1)

  os.environ['CONFIG_DIR_OVERRIDE'] = args.config_dir

  if not common.has_file_in_path('gcloud'):
    print('Please install gcloud.')
    sys.exit(1)

  is_ci = os.getenv('TEST_BOT_ENVIRONMENT')
  if not is_ci and common.is_git_dirty():
    print('Your branch is dirty. Please fix before deploying.')
    sys.exit(1)

  if not common.has_file_in_path('gsutil'):
    print('gsutil not found in PATH.')
    sys.exit(1)

  # Build templates before deployment.
  appengine.build_templates()

  if not is_ci and not args.staging:
    if is_diff_origin_master():
      if args.force:
        print('You are not on origin/master. --force is used. Continue.')
        for _ in range(3):
          print('.')
          time.sleep(1)
        print()
      else:
        print('You are not on origin/master. Please fix or use --force.')
        sys.exit(1)

  if args.staging:
    revision = common.compute_staging_revision()
    platforms = ['linux']  # No other platforms required.
  elif args.prod:
    revision = common.compute_prod_revision()
    platforms = list(constants.PLATFORMS.keys())
  else:
    print('Please specify either --prod or --staging. For production '
          'deployments, you probably want to use deploy.sh from your '
          'configs directory instead.')
    sys.exit(1)

  deploy_zips = 'zips' in args.targets
  deploy_appengine = 'appengine' in args.targets

  is_python3 = sys.version_info.major == 3
  package_zip_paths = []
  if deploy_zips:
    for platform_name in platforms:
      package_zip_paths.append(
          package.package(
              revision, platform_name=platform_name, python3=is_python3))
  else:
    # package.package calls these, so only set these up if we're not packaging,
    # since they can be fairly slow.
    appengine.symlink_dirs()
    common.install_dependencies('linux')
    with open(constants.PACKAGE_TARGET_MANIFEST_PATH, 'w') as f:
      f.write('%s\n' % revision)

  too_large_file_path = find_file_exceeding_limit('src/appengine',
                                                  APPENGINE_FILESIZE_LIMIT)
  if too_large_file_path:
    print(("%s is larger than %d bytes. It wouldn't be deployed to appengine."
           ' Please fix.') % (too_large_file_path, APPENGINE_FILESIZE_LIMIT))
    sys.exit(1)

  if args.staging:
    _staging_deployment_helper(python3=is_python3)
  else:
    _prod_deployment_helper(
        args.config_dir,
        package_zip_paths,
        deploy_appengine,
        python3=is_python3)

  with open(constants.PACKAGE_TARGET_MANIFEST_PATH) as f:
    print('Source updated to %s' % f.read())

  if platforms[-1] != common.get_platform():
    # Make sure the installed dependencies are for the current platform.
    common.install_dependencies()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the Java Script minimizer."""

import unittest

from clusterfuzz._internal.bot.minimizer import js_minimizer
from clusterfuzz._internal.bot.tokenizer import antlr_tokenizer
from clusterfuzz._internal.bot.tokenizer.grammars.JavaScriptLexer import \
    JavaScriptLexer
from clusterfuzz._internal.tests.test_libs import helpers


class JSMinimizerTest(unittest.TestCase):
  """Test for Javascript Minimizer."""

  def setUp(self):
    self._hypotheses_tested = []

    helpers.patch_environ(self)
    helpers.patch(self, [
        'clusterfuzz._internal.bot.minimizer.minimizer.Testcase.prepare_test',
    ])

    self.mock.prepare_test.side_effect = self._mock_prepare_test

    self._tokenizer = antlr_tokenizer.AntlrTokenizer(JavaScriptLexer)
    self._minimizer = js_minimizer.JSMinimizer(
        self._mock_test_function(),
        tokenizer=self._tokenizer.tokenize,
        token_combiner=self._tokenizer.combine)

  def _mock_prepare_test(self, testcase, hypothesis):
    """No need to prepare anything, because tests wont be run. Just save what
      tests would have been run to validate that the correct tokens will be
      removed."""
    text = ''
    for index in hypothesis:
      text += testcase.tokens[index]

    self._hypotheses_tested.append(text)

  def _mock_test_function(self):
    return True

  def test_minimize_empty_string(self):
    """Minimizer does not break on empty data."""
    data = b''

    self._minimizer.minimize(data)

    self.assertEqual(self._hypotheses_tested, [])

  def test_if_hypothesis(self):
    """Test that the minimizer successfully removes all of the if statement
      syntax when it hits a bracket.
      e.g.: if (statement_that_evaluates_to_true) { crash() } -> crash()."""
    data = b'if(boolean) { crash }'

    self._minimizer.minimize(data)

    self.assertIn('if(boolean) {}', self._hypotheses_tested)

  def test_try_catch_hypothesis(self):
    """Test that the minimizer successfully removes all of the try/catch
      syntax when it hits a bracket.
      e.g.: try { crash() } catch(e) {} -> crash()."""
    data = b'try{ crash() } catch(e){ }'
    self._minimizer.minimize(data)

    self.assertIn('try{} catch(e){ }', self._hypotheses_tested)

  def test_handle_if_else(self):
    """Make sure the minimizer runs all of the other hypothesis cleanly on
      if else."""
    data = b'if(boolean) {crash} else { do_something_else }'

    self._minimizer.minimize(data)

    self.assertIn('if(boolean) {}', self._hypotheses_tested)
    self.assertIn('if(boolean) {} else { do_something_else }',
                  self._hypotheses_tested)
    self.assertIn("if(boolean) {crash} else {}", self._hypotheses_tested)

  def test_remove_function_call(self):
    """Test that it removes functions calls effectively."""
    data = b'function name(param1, param2){ stuff inside function }'

    self._minimizer.minimize(data)

    self.assertIn('function name(param1, param2){}', self._hypotheses_tested)

  def test_handle_bracket_with_new_line(self):
    """Test for try/catch with extra whitespace."""
    data = b'try{\n\tcrash()\n}\ncatch(e){\n\n}'

    self._minimizer.minimize(data)

    self.assertIn('try{}', self._hypotheses_tested)
    self.assertIn('try{}\ncatch(e){\n\n}', self._hypotheses_tested)

  def test_handle_function_with_new_lines(self):
    """Test for functions with extra whitespace."""
    data = b'function name(param1,\n\t\tparam2){\n\tstuff inside function\n}'

    self._minimizer.minimize(data)

    self.assertIn('function name(param1,\n\t\tparam2){}',
                  self._hypotheses_tested)

  def test_remove_outer_paren(self):
    """Test that the minimizer successfully removes all of the outer parens
      to check for nested parens.
      e.g.: assertTrue(crash()); -> crash()."""
    data = b'assertTrue(crash());'

    self._minimizer.minimize(data)

    self.assertIn('assertTrue()', self._hypotheses_tested)

  def test_remove_inside_paren(self):
    """Test that minimizer removes everything between the parentheses.
      e.g.: crash(junk, more_junk) -> crash()."""
    data = b'crash(junk, more_junk)'

    self._minimizer.minimize(data)

    self.assertIn('junk, more_junk', self._hypotheses_tested)

  def test_remove_paren_to_start_of_line(self):
    """Tests that the minimizer will remove the whole line (including setting
      vars) when there are parens.
      e.g.: leftover_junk = (function() {
             });."""
    data = b'leftover_junk = (function(){\n})'

    self._minimizer.minimize(data)

    self.assertIn('leftover_junk = (function(){\n})', self._hypotheses_tested)

  def test_remove_paren_with_attached_brackets(self):
    """Test that the minimizer removes the whole line and following brackets
      when there are parens.
      e.g.: (function(global) { })(this);."""
    data = b'(function(global) { })(this)'

    self._minimizer.minimize(data)

    self.assertIn('(function(global) { })(this)', self._hypotheses_tested)

  def test_remove_left_of_comma(self):
    """Test the minimizer removes the comma and token left of the comma
      e.g.: f(whatever, crash()) -> f(crash())."""
    data = b'f(whatever, crash())'

    self._minimizer.minimize(data)

    self.assertIn('whatever,', self._hypotheses_tested)

  def test_remove_right_of_comma(self):
    """Test the minimizer removes the comma and right of the comma.
      e.g.: f(crash(),whatever) -> f(crash())."""
    data = b'f(crash(), whatever)'

    self._minimizer.minimize(data)

    self.assertIn(', whatever', self._hypotheses_tested)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Device information related functions."""

import copy
import datetime
import os
import time

import six

from clusterfuzz._internal.base import dates
from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

from . import adb
from . import app
from . import constants
from . import logger
from . import sanitizer
from . import settings
from . import ui
from . import wifi

# Variables related to adding test account on device.
ADD_TEST_ACCOUNT_APK_NAME = 'user_account_setup.apk'
ADD_TEST_ACCOUNT_CHECK_INTERVAL = 1 * 24 * 60 * 60
ADD_TEST_ACCOUNT_PKG_NAME = 'com.google.android.tests.utilities'
ADD_TEST_ACCOUNT_CALL_PATH = '%s/.AddAccount' % ADD_TEST_ACCOUNT_PKG_NAME
ADD_TEST_ACCOUNT_TIMEOUT = 20

# System build properties related vars.
BUILD_PROP_PATH = '/system/build.prop'
BUILD_PROP_BACKUP_PATH = BUILD_PROP_PATH + '.bak'
BUILD_PROPERTIES = {
    # Disable boot animation.
    'debug.sf.nobootanimation': '1',
    # Disable privileged app permissions enforcement.
    'ro.control_privapp_permissions': 'disable',
    # Scan for wifi less often: saves battery.
    'wifi.supplicant_scan_interval': '500',
}

# Local development settings related vars.
LOCAL_PROP_PATH = '/data/local.prop'
LOCAL_PROP_SETTINGS = [
    'ro.audio.silent=1',
    'ro.monkey=1',
    'ro.setupwizard.mode=DISABLED',
    'ro.test_harness=1',
    'ro.telephony.disable-call=true',
]

# Lockscreen database settings related vars.
LOCKSCREEN_DB = '/data/system/locksettings.db'
LOCKSCREEN_TABLE_NAME = 'locksettings'


def add_test_accounts_if_needed():
  """Add test account to work with GmsCore, etc."""
  last_test_account_check_time = persistent_cache.get_value(
      constants.LAST_TEST_ACCOUNT_CHECK_KEY,
      constructor=datetime.datetime.utcfromtimestamp)
  needs_test_account_update = (
      last_test_account_check_time is None or dates.time_has_expired(
          last_test_account_check_time,
          seconds=ADD_TEST_ACCOUNT_CHECK_INTERVAL))
  if not needs_test_account_update:
    return

  config = db_config.get()
  if not config:
    return

  test_account_email = config.test_account_email
  test_account_password = config.test_account_password
  if not test_account_email or not test_account_password:
    return

  adb.run_as_root()
  wifi.configure(force_enable=True)

  if not app.is_installed(ADD_TEST_ACCOUNT_PKG_NAME):
    logs.log('Installing helper apk for adding test account.')
    android_directory = environment.get_platform_resources_directory()
    add_test_account_apk_path = os.path.join(android_directory,
                                             ADD_TEST_ACCOUNT_APK_NAME)
    app.install(add_test_account_apk_path)

  logs.log('Trying to add test account.')
  output = adb.run_shell_command(
      'am instrument -e account %s -e password %s -w %s' %
      (test_account_email, test_account_password, ADD_TEST_ACCOUNT_CALL_PATH),
      timeout=ADD_TEST_ACCOUNT_TIMEOUT)
  if not output or test_account_email not in output:
    logs.log('Failed to add test account, probably due to wifi issues.')
    return

  logs.log('Test account added successfully.')
  persistent_cache.set_value(constants.LAST_TEST_ACCOUNT_CHECK_KEY, time.time())


def clear_temp_directories():
  """Clear temp directories."""
  adb.remove_directory(constants.DEVICE_DOWNLOAD_DIR, recreate=True)
  adb.remove_directory(constants.DEVICE_TMP_DIR, recreate=True)
  adb.remove_directory(constants.DEVICE_FUZZING_DIR, recreate=True)


def clear_testcase_directory():
  """Clears testcase directory."""
  adb.remove_directory(constants.DEVICE_TESTCASES_DIR, recreate=True)


def configure_device_settings():
  """Configures device settings for test environment."""
  adb.run_as_root()

  # The following line filled with magic numbers will set media volume to 0
  # 3 is the 3rd function in the IAudioServiceList and the following
  # i32's specify 32 bit integer arguments to the function
  adb.run_shell_command('service call audio 3 i32 3 i32 0 i32 1')

  # FIXME: We shouldn't need repeat invocation of this. We need to do this
  # in case previous invocations of any of the below commands failed.
  # Write our test environment settings in content database.
  settings.set_content_setting('com.google.settings/partner',
                               'use_location_for_services', 0)
  settings.set_content_setting('settings/global', 'assisted_gps_enabled', 0)
  settings.set_content_setting('settings/global',
                               'development_settings_enabled', 0)
  settings.set_content_setting('settings/global', 'stay_on_while_plugged_in', 3)
  settings.set_content_setting('settings/global', 'send_action_app_error', 0)
  settings.set_content_setting('settings/global',
                               'verifier_verify_adb_installs', 0)
  settings.set_content_setting('settings/global', 'wifi_scan_always_enabled', 0)
  settings.set_content_setting('settings/secure', 'anr_show_background', 0)
  settings.set_content_setting('settings/secure', 'doze_enabled', 0)
  settings.set_content_setting('settings/secure', 'location_providers_allowed',
                               '')
  settings.set_content_setting('settings/secure', 'lockscreen.disabled', 1)
  settings.set_content_setting('settings/secure', 'screensaver_enabled', 0)
  settings.set_content_setting('settings/system', 'accelerometer_rotation', 0)
  settings.set_content_setting('settings/system', 'auto_time', 0)
  settings.set_content_setting('settings/system', 'auto_timezone', 0)
  settings.set_content_setting('settings/system', 'lockscreen.disabled', 1)
  settings.set_content_setting('settings/system', 'notification_light_pulse', 0)
  settings.set_content_setting('settings/system', 'screen_brightness_mode', 0)
  settings.set_content_setting('settings/system', 'screen_brightness', 1)
  settings.set_content_setting('settings/system', 'user_rotation', 0)

  # On certain device/Android configurations we need to disable the lock screen
  # in a different database. Additionally, the password type must be set to 0.
  settings.set_database_setting(LOCKSCREEN_DB, LOCKSCREEN_TABLE_NAME,
                                'lockscreen.disabled', 1)
  settings.set_database_setting(LOCKSCREEN_DB, LOCKSCREEN_TABLE_NAME,
                                'lockscreen.password_type', 0)
  settings.set_database_setting(LOCKSCREEN_DB, LOCKSCREEN_TABLE_NAME,
                                'lockscreen.password_type_alternate', 0)

  app.disable_packages_that_crash_with_gestures()

  # Create a list of property name and names to be used in local.prop file.
  local_properties_settings_list = copy.deepcopy(LOCAL_PROP_SETTINGS)

  # Add debugging flags to local settings list so that they persist across
  # reboots.
  local_properties_settings_list += get_debug_props_and_values()

  # Write the local properties file settings.
  local_properties_file_contents = '\n'.join(local_properties_settings_list)
  adb.write_data_to_file(local_properties_file_contents, LOCAL_PROP_PATH)


def configure_system_build_properties():
  """Modifies system build properties in /system/build.prop for better boot
  speed and power use."""
  adb.run_as_root()

  # Check md5 checksum of build.prop to see if already updated,
  # in which case exit. If build.prop does not exist, something
  # is very wrong with the device, so bail.
  old_md5 = persistent_cache.get_value(constants.BUILD_PROP_MD5_KEY)
  current_md5 = adb.get_file_checksum(BUILD_PROP_PATH)
  if current_md5 is None:
    logs.log_error('Unable to find %s on device.' % BUILD_PROP_PATH)
    return
  if old_md5 == current_md5:
    return

  # Pull to tmp file.
  bot_tmp_directory = environment.get_value('BOT_TMPDIR')
  old_build_prop_path = os.path.join(bot_tmp_directory, 'old.prop')
  adb.run_command(['pull', BUILD_PROP_PATH, old_build_prop_path])
  if not os.path.exists(old_build_prop_path):
    logs.log_error('Unable to fetch %s from device.' % BUILD_PROP_PATH)
    return

  # Write new build.prop.
  new_build_prop_path = os.path.join(bot_tmp_directory, 'new.prop')
  old_build_prop_file_content = open(old_build_prop_path, 'r')
  new_build_prop_file_content = open(new_build_prop_path, 'w')
  new_content_notification = '### CHANGED OR ADDED PROPERTIES ###'
  for line in old_build_prop_file_content:
    property_name = line.split('=')[0].strip()
    if property_name in BUILD_PROPERTIES:
      continue
    if new_content_notification in line:
      continue
    new_build_prop_file_content.write(line)

  new_build_prop_file_content.write(new_content_notification + '\n')
  for flag, value in six.iteritems(BUILD_PROPERTIES):
    new_build_prop_file_content.write('%s=%s\n' % (flag, value))
  old_build_prop_file_content.close()
  new_build_prop_file_content.close()

  # Keep verified boot disabled for M and higher releases. This makes it easy
  # to modify system's app_process to load asan libraries.
  build_version = settings.get_build_version()
  if is_build_at_least(build_version, 'M'):
    adb.run_as_root()
    adb.run_command('disable-verity')
    reboot()

  # Make /system writable.
  adb.run_as_root()
  adb.remount()

  # Remove seccomp policies (on N and higher) as ASan requires extra syscalls.
  if is_build_at_least(build_version, 'N'):
    policy_files = adb.run_shell_command(
        ['find', '/system/etc/seccomp_policy/', '-type', 'f'])
    for policy_file in policy_files.splitlines():
      adb.run_shell_command(['rm', policy_file.strip()])

  # Push new build.prop and backup to device.
  logs.log('Pushing new build properties file on device.')
  adb.run_command(['push', '-p', old_build_prop_path, BUILD_PROP_BACKUP_PATH])
  adb.run_command(['push', '-p', new_build_prop_path, BUILD_PROP_PATH])
  adb.run_shell_command(['chmod', '644', BUILD_PROP_PATH])

  # Set persistent cache key containing and md5sum.
  current_md5 = adb.get_file_checksum(BUILD_PROP_PATH)
  persistent_cache.set_value(constants.BUILD_PROP_MD5_KEY, current_md5)


def get_debug_props_and_values():
  """Return debug property names and values based on |ENABLE_DEBUG_CHECKS|
  flag."""
  debug_props_and_values_list = []
  enable_debug_checks = environment.get_value('ENABLE_DEBUG_CHECKS', False)

  logs.log('Debug flags set to %s.' % str(enable_debug_checks))

  # Keep system and applications level asserts disabled since these can lead to
  # potential battery depletion issues.
  debug_props_and_values_list += [
      'dalvik.vm.enableassertions=',
      'debug.assert=0',
  ]

  # JNI checks. See this link for more information.
  # http://android-developers.blogspot.com/2011/07/debugging-android-jni-with-checkjni.html.
  check_jni_flag = (
      enable_debug_checks or environment.get_value('ENABLE_CHECK_JNI', False))
  debug_props_and_values_list += [
      'dalvik.vm.checkjni=%s' % str(check_jni_flag).lower(),
      'debug.checkjni=%d' % int(check_jni_flag),
  ]

  is_build_supported = is_build_at_least(settings.get_build_version(), 'N')
  debug_malloc_enabled = (
      enable_debug_checks and is_build_supported and
      not settings.get_sanitizer_tool_name())

  # https://android.googlesource.com/platform/bionic/+/master/libc/malloc_debug/README.md
  if debug_malloc_enabled:
    # FIXME: 'free_track' is very crashy. Skip for now.
    debug_malloc_string = 'fill guard'
    debug_props_and_values_list += [
        'libc.debug.malloc.options=%s' % debug_malloc_string
    ]

  return debug_props_and_values_list


def initialize_device():
  """Prepares android device for app install."""
  if environment.is_engine_fuzzer_job() or environment.is_kernel_fuzzer_job():
    # These steps are not applicable to libFuzzer and syzkaller jobs and can
    # brick a device on trying to configure device build settings.
    return

  adb.setup_adb()

  # General device configuration settings.
  configure_system_build_properties()
  configure_device_settings()
  add_test_accounts_if_needed()

  # Setup AddressSanitizer if needed.
  sanitizer.setup_asan_if_needed()

  # Reboot device as above steps would need it and also it brings device in a
  # good state.
  reboot()

  # Make sure we are running as root after restart.
  adb.run_as_root()

  # Other configuration tasks (only to done after reboot).
  wifi.configure()
  setup_host_and_device_forwarder_if_needed()
  settings.change_se_linux_to_permissive_mode()
  app.wait_until_optimization_complete()
  ui.clear_notifications()
  ui.unlock_screen()

  # FIXME: Should we should revert back to regular user permission ?


def initialize_environment():
  """Set common environment variables for easy access."""
  environment.set_value('BUILD_FINGERPRINT', settings.get_build_fingerprint())
  environment.set_value('BUILD_VERSION', settings.get_build_version())
  environment.set_value('DEVICE_CODENAME', settings.get_device_codename())
  environment.set_value('DEVICE_PATH', adb.get_device_path())
  environment.set_value('PLATFORM_ID', settings.get_platform_id())
  environment.set_value('PRODUCT_BRAND', settings.get_product_brand())
  environment.set_value('SANITIZER_TOOL_NAME',
                        settings.get_sanitizer_tool_name())
  environment.set_value('SECURITY_PATCH_LEVEL',
                        settings.get_security_patch_level())


def install_application_if_needed(apk_path, force_update):
  """Install application package if it does not exist on device
  or if force_update is set."""
  # Make sure that apk exists and has non-zero size. Otherwise, it means we
  # are using a system package that we just want to fuzz, but not care about
  # installation.
  if (not apk_path or not os.path.exists(apk_path) or
      not os.path.getsize(apk_path)):
    return

  # If we don't have a package name, we can't uninstall the app. This is needed
  # for installation workflow.
  package_name = app.get_package_name()
  if not package_name:
    return

  # Add |REINSTALL_APP_BEFORE_EACH_TASK| to force update decision.
  reinstall_app_before_each_task = environment.get_value(
      'REINSTALL_APP_BEFORE_EACH_TASK', False)
  force_update = force_update or reinstall_app_before_each_task

  # Install application if it is not found in the device's
  # package list or force_update flag has been set.
  if force_update or not app.is_installed(package_name):
    app.uninstall(package_name)
    app.install(apk_path)

    if not app.is_installed(package_name):
      logs.log_error(
          'Package %s was not installed successfully.' % package_name)
      return

    logs.log('Package %s is successfully installed using apk %s.' %
             (package_name, apk_path))

  app.reset()


def is_build_at_least(current_version, other_version):
  """Returns whether or not |current_version| is at least as new as
  |other_version|."""
  if current_version is None:
    return False

  # Special-cases for master builds.
  if current_version == 'MASTER':
    # If the current build is master, we consider it at least as new as any
    # other.
    return True

  if other_version == 'MASTER':
    # Since this build is not master, it is not at least as new as master.
    return False

  return current_version >= other_version


def push_testcases_to_device():
  """Pushes testcases from local fuzz directory onto device."""
  # Attempt to ensure that the local state is the same as the state on the
  # device by clearing existing files on device before pushing.
  clear_testcase_directory()

  local_testcases_directory = environment.get_value('FUZZ_INPUTS')
  if not os.listdir(local_testcases_directory):
    # Directory is empty, nothing to push.
    logs.log('No testcases to copy to device, skipping.')
    return

  adb.copy_local_directory_to_remote(local_testcases_directory,
                                     constants.DEVICE_TESTCASES_DIR)


def reboot():
  """Reboots device and clear config state."""
  # Make sure to clear logcat before reboot occurs. In case of kernel crashes,
  # we use the log before reboot, so it is good to clear it when we are doing
  # the reboot explicitly.
  logger.clear_log()

  # Reboot.
  logs.log('Rebooting device.')
  adb.reboot()

  # Wait for boot to complete.
  adb.wait_until_fully_booted()


def setup_host_and_device_forwarder_if_needed():
  """Sets up http(s) forwarding between device and host."""
  # Get list of ports to map.
  http_port_1 = environment.get_value('HTTP_PORT_1', 8000)
  http_port_2 = environment.get_value('HTTP_PORT_2', 8080)
  ports = [http_port_1, http_port_2]

  # Reverse map socket connections from device to host machine.
  for port in ports:
    port_string = 'tcp:%d' % port
    adb.run_command(['reverse', port_string, port_string])


def update_build(apk_path, force_update=True, should_initialize_device=True):
  """Prepares the device and updates the build if necessary."""
  # Prepare device for app install.
  if should_initialize_device:
    initialize_device()

  # On Android, we may need to write a command line file. We do this in
  # advance so that we do not have to write this to the device multiple
  # times.
  # TODO(mbarbella): Platforms code should not depend on bot.
  from clusterfuzz._internal.bot import testcase_manager
  testcase_manager.get_command_line_for_application(
      write_command_line_file=True)

  # Install the app if it does not exist.
  install_application_if_needed(apk_path, force_update=force_update)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for crash uploader functions."""

import os
import socket
import unittest
import urllib.parse

import mock

from clusterfuzz._internal.chrome import crash_uploader
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_analyzer
from clusterfuzz._internal.protos import process_state_pb2
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils

DATA_DIRECTORY = os.path.join(os.path.dirname(__file__), 'crash_uploader_data')

# CrashReportInfo test variables.
# On crash, Android devices store a MIME containing the crash minidump as well
# as other information needed for chromecrash upload. This file is written to a
# .dmp file, suffixed by the PID to distinguish it from the minidump file
# itself.
SAMPLE_MIME_FILENAME = 'android_mime_minidump'
SAMPLE_MIME_PATH = os.path.join(DATA_DIRECTORY,
                                '%s.mime' % SAMPLE_MIME_FILENAME)
EXPECTED_DMP_PATH = os.path.join(DATA_DIRECTORY,
                                 'android_parsed_minidump_expected.dmp')
ACTUAL_DMP_PATH = os.path.join(DATA_DIRECTORY, '%s.dmp' % SAMPLE_MIME_FILENAME)
EXPECTED_PROCESSED_REPORT_PATH = os.path.join(
    DATA_DIRECTORY, 'expected_processed_report_bytes')
SAMPLE_OUTPUT = open(
    os.path.join(DATA_DIRECTORY, 'android_crash_stack_output'), 'r').read()
SAMPLE_OUTPUT_TO_PARSE = open(
    os.path.join(DATA_DIRECTORY, 'crash_output_to_parse'), 'r').read()
EXPECTED_REPORT_INFO = crash_uploader.CrashReportInfo(
    minidump_path=ACTUAL_DMP_PATH,
    product='Chrome_Android',
    version='46.0.2482.0')

# Environment variables.
TEST_JOB_NAME = 'android_asan_chrome_l'
TEST_OS = 'android'
TEST_FAIL_RETRIES = '4'
TEST_BOT_TMPDIR = DATA_DIRECTORY
TEST_CRASH_STACKTRACES_DIR = DATA_DIRECTORY


class CrashBaseTest(unittest.TestCase):
  """Base for setup, teardown of crash report processing tests."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.crash_analysis.stack_parsing.stack_symbolizer.filter_binary_path'
    ])

    # For Android, skip symbolization as real device is unavailable.
    self.mock.filter_binary_path.return_value = ''

    os.environ['APP_NAME'] = 'chrome'
    os.environ['BOT_TMPDIR'] = TEST_BOT_TMPDIR
    os.environ['CRASH_STACKTRACES_DIR'] = TEST_CRASH_STACKTRACES_DIR
    os.environ['FAIL_RETRIES'] = TEST_FAIL_RETRIES
    os.environ['JOB_NAME'] = TEST_JOB_NAME
    os.environ['OS_OVERRIDE'] = TEST_OS
    os.environ['UPLOAD_MODE'] = 'staging'

  def tearDown(self):
    if self.needs_file_delete:
      os.remove(ACTUAL_DMP_PATH)


class CrashReportsTest(CrashBaseTest):
  """Crash report processing tests."""

  def _validate_dmp_contents(self, my_expected_dmp_path, my_actual_dmp_path):
    """Check dmp contents at given path are as expected."""
    try:
      expected_dmp_file_contents = open(my_expected_dmp_path, 'rb').read()
      actual_dmp_file_contents = open(my_actual_dmp_path, 'rb').read()
      self.assertEqual(actual_dmp_file_contents, expected_dmp_file_contents)
    except IOError:
      self.fail('Could not find processed %s or %s.' % (my_expected_dmp_path,
                                                        my_actual_dmp_path))

  def _validate_report_fields(self, my_expected_report_info,
                              my_actual_report_info):
    """Check CrashReportInfo object created correctly."""
    # Only check fields the uploader will need (e.g. we don't care so much
    # about un/symbolized stacktraces.
    self.assertEqual(my_actual_report_info.minidump_info.path,
                     my_expected_report_info.minidump_info.path)
    self.assertEqual(my_actual_report_info.serialized_crash_stack_frames,
                     my_expected_report_info.serialized_crash_stack_frames)
    self.assertEqual(my_actual_report_info.product,
                     my_expected_report_info.product)
    self.assertEqual(my_actual_report_info.version,
                     my_expected_report_info.version)

  @test_utils.supported_platforms('LINUX')
  def test_parse_output_to_processed_report(self):
    """Tests if given output parses to the expected symbolized stack bytes."""
    self.needs_file_delete = False
    state = stack_analyzer.get_crash_data(SAMPLE_OUTPUT_TO_PARSE)
    actual_report_bytes = crash_uploader.get_symbolized_stack_bytes(
        state.crash_type, state.crash_address, state.frames)
    with open(EXPECTED_PROCESSED_REPORT_PATH, 'rb') as expected_report:
      expected_report_bytes = expected_report.read()

    self.assertEqual(actual_report_bytes, expected_report_bytes)

  def test_parse_output_to_processed_report_bad_stack(self):
    """Tests if given bad stack fails to parse to a processed report."""
    self.needs_file_delete = False
    actual_report_bytes = crash_uploader.get_symbolized_stack_bytes(
        'reason', 12345, [None])
    self.assertIsNone(actual_report_bytes)

  def test_parse_output_to_processed_report_big_addresses(self):
    """Tests that given output with big addresses prototizes."""
    self.needs_file_delete = False
    actual_report_bytes = crash_uploader.get_symbolized_stack_bytes(
        'reason', '0xfffffffffffffff4', [])
    actual_report_proto = process_state_pb2.ProcessStateProto()
    actual_report_proto.ParseFromString(actual_report_bytes)

    self.assertEqual(actual_report_proto.crash.address, -12)

  def test_to_report_metadata_and_back(self):
    """Test to report metadata and back."""
    self.needs_file_delete = False

    expected_report_info = crash_uploader.CrashReportInfo()
    expected_report_info.product = 'Chrome_Android'
    expected_report_info.version = '59.0.3035.0'
    expected_report_info.testcase_id = 'CF_TESTSUITE_TEST_UPLOAD'
    expected_report_info.bot_id = 'test_upload'
    with open(EXPECTED_PROCESSED_REPORT_PATH, 'rb') as processed_report:
      expected_report_info.serialized_crash_stack_frames = (
          processed_report.read())

    report_metadata = expected_report_info.to_report_metadata()
    actual_report_info = crash_uploader.crash_report_info_from_metadata(
        report_metadata)

    self._validate_report_fields(expected_report_info, actual_report_info)

  def test_parse_mime_to_crash_report_info(self):
    """Tests if parsing sample MIME file produces expected CrashReportInfo
       object."""
    self.needs_file_delete = True
    report_info = crash_uploader.parse_mime_to_crash_report_info(
        SAMPLE_MIME_PATH)

    # Check processed (extracted) dmp contents are what we expect.
    self._validate_dmp_contents(EXPECTED_DMP_PATH, ACTUAL_DMP_PATH)

    # Check object fields are what we expect.
    self._validate_report_fields(EXPECTED_REPORT_INFO, report_info)

  @mock.patch('clusterfuzz._internal.platforms.android.adb.run_shell_command')
  @mock.patch('clusterfuzz._internal.platforms.android.adb.run_command')
  def test_get_crash_info(self, mock_run_shell_command, mock_run_command):
    """Tests if parsing sample output (crash-stacks stacktrace) produces
       expected CrashReportInfo object (on Android; on other platforms,
       should fail)."""
    self.needs_file_delete = True

    # Process the output string, mocking the adb commands appropriately.
    sample_mime_device_path = os.path.join(
        '/data/data/com.google.android.apps.chrome/cache/Crash Reports',
        '%s.dmp1' % SAMPLE_MIME_FILENAME)
    mock_run_shell_command.return_value = sample_mime_device_path
    mock_run_command.return_value = sample_mime_device_path
    report_info = crash_uploader.get_crash_info(SAMPLE_OUTPUT)

    # Check processed (extracted) dmp contents are what we expect.
    self._validate_dmp_contents(EXPECTED_DMP_PATH, ACTUAL_DMP_PATH)

    # Check object fields are what we expect.
    self._validate_report_fields(EXPECTED_REPORT_INFO, report_info)

    # Sanity checks for failing on other platforms.
    os.environ['OS_OVERRIDE'] = 'mac'
    if crash_uploader.get_crash_info(SAMPLE_OUTPUT):
      os.environ['OS_OVERRIDE'] = TEST_OS
      self.fail('Expected none for non-Android.')

  @mock.patch('clusterfuzz._internal.google_cloud_utils.blobs.write_blob')
  def test_store_minidump(self, mock_write_testcase):
    """Tests (very roughly) minidump upload to blobstore: just check there /is/
       a blobstore ID returned."""
    self.needs_file_delete = False
    mock_write_testcase.return_value = '11111'

    sample_report_info = crash_uploader.CrashReportInfo(
        minidump_path=EXPECTED_DMP_PATH,
        product='Chrome_Android',
        version='46.0.2482.0')
    minidump_key = sample_report_info.store_minidump()
    if not minidump_key:
      self.fail('Could not upload minidump to blobstore.')

  def test_get_crash_info_and_stacktrace(self):
    """Tests crash stacktrace retry logic."""
    self.needs_file_delete = False

    with mock.patch.object(crash_uploader,
                           'get_crash_info') as mock_get_crash_info:
      with mock.patch.object(process_handler,
                             'run_process') as mock_run_process:
        # First try expected retry success.
        mock_get_crash_info.side_effect = [None, EXPECTED_REPORT_INFO]
        mock_run_process.side_effect = ([(None, None, SAMPLE_OUTPUT)] * 2)
        crash_info, _ = crash_uploader.get_crash_info_and_stacktrace(
            '', SAMPLE_OUTPUT, None)
        if crash_info is None:
          self.fail('Could not get crash info after retries.')

        # And check expected retry failure.
        retry_limit = int(os.environ['FAIL_RETRIES'])
        mock_get_crash_info.side_effect = [None] * (retry_limit + 1)
        mock_run_process.side_effect = (
            [(None, None, SAMPLE_OUTPUT)] * (retry_limit + 1))
        crash_info, _ = crash_uploader.get_crash_info_and_stacktrace(
            '', SAMPLE_OUTPUT, None)
        if crash_info is not None:
          self.fail('Expected failure to get crash info after retries, '
                    'got success.')


@test_utils.integration
class CrashUploadTest(CrashBaseTest):
  """Crash upload tests."""

  @classmethod
  def setUpClass(cls):
    os.environ['UPLOAD_MODE'] = 'staging'
    cls.server_error = cls.get_server_error()

  @classmethod
  def get_server_error(cls):
    """Get server error."""
    upload_url = crash_uploader.CRASH_REPORT_UPLOAD_URL[environment.get_value(
        'UPLOAD_MODE')]
    ping_url = urllib.parse.urlsplit(upload_url).netloc
    try:
      # Use a port that has been used for crash/ uploads before.
      sock = socket.create_connection((ping_url, 443), timeout=1)
      sock.close()
      return None
    except (socket.error, socket.timeout) as e:
      return 'Failed to connect to crash/: %s' % e

  def test_upload_processed_report(self):
    """Tests (very roughly) crash report upload with a processed report: just
       check there /is/ a returned report id."""
    self.needs_file_delete = False

    report_info = crash_uploader.CrashReportInfo()
    report_info.product = 'Chrome_Android'
    report_info.version = '59.0.3035.0'
    report_info.testcase_id = 'CF_TESTSUITE_TEST_UPLOAD'
    report_info.bot_id = 'test_upload'
    with open(EXPECTED_PROCESSED_REPORT_PATH, 'rb') as processed_report:
      report_info.serialized_crash_stack_frames = processed_report.read()

    # Attempt upload.
    if self.server_error:
      self.fail(self.server_error)
    report_id = report_info.upload()
    if not report_id:
      self.fail('No report id returned.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolize task.
   Add stack traces from non-optimized release and debug builds."""

import os

from clusterfuzz._internal.base import tasks
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.tasks import setup
from clusterfuzz._internal.bot.tasks import task_creation
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.crash_analysis import crash_analyzer
from clusterfuzz._internal.crash_analysis.crash_result import CrashResult
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import process_handler

DEFAULT_REDZONE = 128
MAX_REDZONE = 1024
MIN_REDZONE = 16
STACK_FRAME_COUNT = 128


def execute_task(testcase_id, job_type):
  """Execute a symbolize command."""
  # Locate the testcase associated with the id.
  testcase = data_handler.get_testcase_by_id(testcase_id)

  # We should atleast have a symbolized debug or release build.
  if not build_manager.has_symbolized_builds():
    return

  data_handler.update_testcase_comment(testcase, data_types.TaskState.STARTED)

  # Setup testcase and its dependencies.
  file_list, _, testcase_file_path = setup.setup_testcase(testcase, job_type)
  if not file_list:
    return

  # Initialize variables.
  build_fail_wait = environment.get_value('FAIL_WAIT')

  old_crash_stacktrace = data_handler.get_stacktrace(testcase)
  sym_crash_type = testcase.crash_type
  sym_crash_address = testcase.crash_address
  sym_crash_state = testcase.crash_state
  sym_redzone = DEFAULT_REDZONE
  warmup_timeout = environment.get_value('WARMUP_TIMEOUT')

  # Decide which build revision to use.
  if testcase.crash_stacktrace == 'Pending':
    # This usually happen when someone clicked the 'Update stacktrace from
    # trunk' button on the testcase details page. In this case, we are forced
    # to use trunk. No revision -> trunk build.
    build_revision = None
  else:
    build_revision = testcase.crash_revision

  # Set up a custom or regular build based on revision.
  build_manager.setup_build(build_revision)

  # Get crash revision used in setting up build.
  crash_revision = environment.get_value('APP_REVISION')

  if not build_manager.check_app_path():
    testcase = data_handler.get_testcase_by_id(testcase_id)
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         'Build setup failed')
    tasks.add_task(
        'symbolize', testcase_id, job_type, wait_time=build_fail_wait)
    return

  # ASAN tool settings (if the tool is used).
  # See if we can get better stacks with higher redzone sizes.
  # A UAF might actually turn out to be OOB read/write with a bigger redzone.
  if environment.tool_matches('ASAN', job_type) and testcase.security_flag:
    redzone = MAX_REDZONE
    while redzone >= MIN_REDZONE:
      environment.reset_current_memory_tool_options(
          redzone_size=testcase.redzone, disable_ubsan=testcase.disable_ubsan)

      process_handler.terminate_stale_application_instances()
      command = testcase_manager.get_command_line_for_application(
          testcase_file_path, needs_http=testcase.http_flag)
      return_code, crash_time, output = (
          process_handler.run_process(
              command, timeout=warmup_timeout, gestures=testcase.gestures))
      crash_result = CrashResult(return_code, crash_time, output)

      if crash_result.is_crash() and 'AddressSanitizer' in output:
        state = crash_result.get_symbolized_data()
        security_flag = crash_result.is_security_issue()

        if (not crash_analyzer.ignore_stacktrace(state.crash_stacktrace) and
            security_flag == testcase.security_flag and
            state.crash_type == testcase.crash_type and
            (state.crash_type != sym_crash_type or
             state.crash_state != sym_crash_state)):
          logs.log('Changing crash parameters.\nOld : %s, %s, %s' %
                   (sym_crash_type, sym_crash_address, sym_crash_state))

          sym_crash_type = state.crash_type
          sym_crash_address = state.crash_address
          sym_crash_state = state.crash_state
          sym_redzone = redzone
          old_crash_stacktrace = state.crash_stacktrace

          logs.log('\nNew : %s, %s, %s' % (sym_crash_type, sym_crash_address,
                                           sym_crash_state))
          break

      redzone /= 2

  # We should have atleast a symbolized debug or a release build.
  symbolized_builds = build_manager.setup_symbolized_builds(crash_revision)
  if (not symbolized_builds or
      (not build_manager.check_app_path() and
       not build_manager.check_app_path('APP_PATH_DEBUG'))):
    testcase = data_handler.get_testcase_by_id(testcase_id)
    data_handler.update_testcase_comment(testcase, data_types.TaskState.ERROR,
                                         'Build setup failed')
    tasks.add_task(
        'symbolize', testcase_id, job_type, wait_time=build_fail_wait)
    return

  # Increase malloc_context_size to get all stack frames. Default is 30.
  environment.reset_current_memory_tool_options(
      redzone_size=sym_redzone,
      malloc_context_size=STACK_FRAME_COUNT,
      symbolize_inline_frames=True,
      disable_ubsan=testcase.disable_ubsan)

  # TSAN tool settings (if the tool is used).
  if environment.tool_matches('TSAN', job_type):
    environment.set_tsan_max_history_size()

  # Do the symbolization if supported by this application.
  result, sym_crash_stacktrace = (
      get_symbolized_stacktraces(testcase_file_path, testcase,
                                 old_crash_stacktrace, sym_crash_state))

  # Update crash parameters.
  testcase = data_handler.get_testcase_by_id(testcase_id)
  testcase.crash_type = sym_crash_type
  testcase.crash_address = sym_crash_address
  testcase.crash_state = sym_crash_state
  testcase.crash_stacktrace = (
      data_handler.filter_stacktrace(sym_crash_stacktrace))

  if not result:
    data_handler.update_testcase_comment(
        testcase, data_types.TaskState.ERROR,
        'Unable to reproduce crash, skipping '
        'stacktrace update')
  else:
    # Switch build url to use the less-optimized symbolized build with better
    # stacktrace.
    build_url = environment.get_value('BUILD_URL')
    if build_url:
      testcase.set_metadata('build_url', build_url, update_testcase=False)

    data_handler.update_testcase_comment(testcase,
                                         data_types.TaskState.FINISHED)

  testcase.symbolized = True
  testcase.crash_revision = crash_revision
  testcase.put()

  # We might have updated the crash state. See if we need to marked as duplicate
  # based on other testcases.
  data_handler.handle_duplicate_entry(testcase)

  task_creation.create_blame_task_if_needed(testcase)

  # Switch current directory before builds cleanup.
  root_directory = environment.get_value('ROOT_DIR')
  os.chdir(root_directory)

  # Cleanup symbolized builds which are space-heavy.
  symbolized_builds.delete()


def get_symbolized_stacktraces(testcase_file_path, testcase,
                               old_crash_stacktrace, expected_state):
  """Use the symbolized builds to generate an updated stacktrace."""
  # Initialize variables.
  app_path = environment.get_value('APP_PATH')
  app_path_debug = environment.get_value('APP_PATH_DEBUG')
  long_test_timeout = environment.get_value('WARMUP_TIMEOUT')
  retry_limit = environment.get_value('FAIL_RETRIES')
  symbolized = False

  debug_build_stacktrace = ''
  release_build_stacktrace = old_crash_stacktrace

  # Symbolize using the debug build first so that the debug build stacktrace
  # comes after the more important release build stacktrace.
  if app_path_debug:
    for _ in range(retry_limit):
      process_handler.terminate_stale_application_instances()
      command = testcase_manager.get_command_line_for_application(
          testcase_file_path,
          app_path=app_path_debug,
          needs_http=testcase.http_flag)
      return_code, crash_time, output = (
          process_handler.run_process(
              command, timeout=long_test_timeout, gestures=testcase.gestures))
      crash_result = CrashResult(return_code, crash_time, output)

      if crash_result.is_crash():
        state = crash_result.get_symbolized_data()

        if crash_analyzer.ignore_stacktrace(state.crash_stacktrace):
          continue

        unsymbolized_crash_stacktrace = crash_result.get_stacktrace(
            symbolized=False)
        debug_build_stacktrace = utils.get_crash_stacktrace_output(
            command,
            state.crash_stacktrace,
            unsymbolized_crash_stacktrace,
            build_type='debug')
        symbolized = True
        break

  # Symbolize using the release build.
  if app_path:
    for _ in range(retry_limit):
      process_handler.terminate_stale_application_instances()
      command = testcase_manager.get_command_line_for_application(
          testcase_file_path, app_path=app_path, needs_http=testcase.http_flag)
      return_code, crash_time, output = (
          process_handler.run_process(
              command, timeout=long_test_timeout, gestures=testcase.gestures))
      crash_result = CrashResult(return_code, crash_time, output)

      if crash_result.is_crash():
        state = crash_result.get_symbolized_data()

        if crash_analyzer.ignore_stacktrace(state.crash_stacktrace):
          continue

        if state.crash_state != expected_state:
          continue

        # Release stack's security flag has to match the symbolized release
        # stack's security flag.
        security_flag = crash_result.is_security_issue()
        if security_flag != testcase.security_flag:
          continue

        unsymbolized_crash_stacktrace = crash_result.get_stacktrace(
            symbolized=False)
        release_build_stacktrace = utils.get_crash_stacktrace_output(
            command,
            state.crash_stacktrace,
            unsymbolized_crash_stacktrace,
            build_type='release')
        symbolized = True
        break

  stacktrace = release_build_stacktrace
  if debug_build_stacktrace:
    stacktrace += '\n\n' + debug_build_stacktrace

  return symbolized, stacktrace



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the stack analyzer module for lkl specifically."""

import os
import unittest

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.crash_analysis.stack_parsing import stack_analyzer
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers

DATA_DIRECTORY = os.path.join(os.path.dirname(__file__), 'stack_analyzer_data')
TEST_JOB_NAME = 'test_lkl'

KERNEL_REPRO = """kernel/build u'6eeca0ea35da517952643a3b0c5b2436df4d3230'
kernel/common-patches u'34bbf764ca33bd39922a40ca641139a9362a0812'
kernel/configs u'89ba434b0b7624d6085f22742f01a0d8502e7025'
kernel/cuttlefish-modules u'5375d668eb898993d0acbba27de47e2eb605b0a0'
kernel/goldfish-modules u'f21b72a86766714b06bdc2b532ebd61b10286b26'
kernel/hikey-modules u'33d4b7f884500689a297df156b1c0a29e1c525da'
kernel/manifest u'c85b8efb8d71e931b400942f7716bb5ecac9b1fa'
kernel/prebuilts/build-tools u'96755e2ffffe500011eecd81b63eeaf2d484338e'
kernel/private/lkl u'd0fcd2ee3504f53bba1227805a8ba3828e9279aa'
kernel/tests u'9e155c5f1646097bfa3e71017079b1caf1ba57d4'
platform/prebuilts/boot-artifacts u'45d30d07ad284018481ebb419b5d01b5de72ed02'
platform/prebuilts/build-tools u'167903bfd32b60bcff841422710b7cf489c84fec'
platform/prebuilts/clang/host/linux-x86 u'c1bd0e5040ec38682f101fa6ce35e4e2c0079c0e'
platform/prebuilts/gcc/linux-x86/aarch64/aarch64-linux-android-4.9 u'e9c7c9eb5c3ab5d4f1f09b3ce97498fc59c3bdcd'
platform/prebuilts/gcc/linux-x86/arm/arm-linux-androideabi-4.9 u'1b12660791807c225dc682addaa37d0b9468349c'
platform/prebuilts/gcc/linux-x86/host/x86_64-linux-glibc2.17-4.8 u'71ba8516fe4039ddb00dd0976c211d28d3ff8913'
platform/prebuilts/gcc/linux-x86/x86/x86_64-linux-android-4.9 u'337ac2199f94c781e3c97baf2a0027d004d0097f'
platform/system/tools/mkbootimg u'f59f25d67f55ea6ee879678cfa1bfdd0be5f4019'
"""


# pylint: disable=unused-argument
def _mock_symbolize_stacktrace(stacktrace, enable_inline_frames=True):
  """No-op mocked version of symbolize_stacktrace."""
  return stacktrace


def _mock_fetch_artifact_get(bid,
                             target,
                             regex,
                             output_directory,
                             output_filename_override=None):
  if output_filename_override:
    artifact_path = os.path.join(output_directory, output_filename_override)
    with open(artifact_path, 'w') as artifact_file:
      artifact_file.write(KERNEL_REPRO)


# pylint: enable=unused-argument


class LKLStackAnalyzerTest(unittest.TestCase):
  """LKL specific Stack analyzer tests."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.crash_analysis.stack_parsing.stack_symbolizer.'
        'symbolize_stacktrace',
        'clusterfuzz._internal.metrics.logs.log_error',
    ])
    os.environ['JOB_NAME'] = TEST_JOB_NAME

    self.mock.symbolize_stacktrace.side_effect = _mock_symbolize_stacktrace

  def _mock_read_data_from_file(self, file_path, eval_data=True, default=None):
    if file_path.endswith('repo.prop'):
      return self._real_read_data_from_file(file_path, eval_data, default)

    return None

  def _read_test_data(self, name):
    """Helper function to read test data."""
    with open(os.path.join(DATA_DIRECTORY, name)) as handle:
      return handle.read()

  def test_lkl_linkification(self):
    """Test lkl linkification."""
    environment.set_bot_environment()
    self._real_read_data_from_file = utils.read_data_from_file
    test_helpers.patch(self, [
        'clusterfuzz._internal.platforms.android.fetch_artifact.get',
        'clusterfuzz._internal.google_cloud_utils.storage.get_file_from_cache_if_exists',
        'clusterfuzz._internal.google_cloud_utils.storage.store_file_in_cache',
        'clusterfuzz._internal.base.utils.write_data_to_file',
        'clusterfuzz._internal.base.utils.read_data_from_file'
    ])
    self.mock.get.side_effect = _mock_fetch_artifact_get
    self.mock.get_file_from_cache_if_exists.return_value = False
    self.mock.store_file_in_cache.return_value = None
    self.mock.write_data_to_file = None
    self.mock.read_data_from_file.side_effect = self._mock_read_data_from_file

    data = self._read_test_data('lkl_libfuzzer_symbolized.txt')
    expected_stack = self._read_test_data(
        'lkl_libfuzzer_symbolized_linkified.txt')
    actual_state = stack_analyzer.get_crash_data(data)
    self.assertEqual(actual_state.crash_stacktrace, expected_stack)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dictionary to keep track of pit information."""

import os

from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment


def get_path(grammar):
  """Return the path of the peach pit for the given grammar. Return None if the
  Pit does not exist or the grammar is None."""

  pit_dir = os.path.join(environment.get_platform_resources_directory(),
                         'peach', 'pits')
  pit_path = os.path.join(pit_dir, grammar + '.xml')

  if not os.path.exists(pit_path):
    logs.log_error(
        'Pit file for "%s" grammar is not found.' % grammar, pit_path=pit_path)
    return None

  return pit_path
